// Welcome to
// __________         __    __  .__                               __
// \______   \_____ _/  |__/  |_|  |   ____   ______ ____ _____  |  | __ ____
//  |    |  _/\__  \\   __\   __\  | _/ __ \ /  ___//    \\__  \ |  |/ // __ \
//  |    |   \ / __ \|  |  |  | |  |_\  ___/ \___ \|   |  \/ __ \|    <\  ___/
//  |________/(______/__|  |__| |____/\_____>______>___|__(______/__|__\\_____>
//
// This file can be a nice home for your Battlesnake logic and helper functions.
//
// To get you started we've included code to prevent your Battlesnake from moving backwards.
// For more info see docs.battlesnake.com

use core::f64;
use std::collections::BTreeMap;

use log::info;
use ordered_float::OrderedFloat;
use serde_json::{json, Value};

use crate::{
    simulation::{self, SnakeToken},
    Action, Battlesnake, Board, Direction, Game, MAX_HEALTH,
};

impl Battlesnake {
    fn possible_actions_without_heads<'a>(
        &'a self,
        game: &'a Game,
        board: &'a Board,
    ) -> impl Iterator<Item = Direction> + 'a {
        enum_iterator::all::<Direction>()
            .filter(|direction| {
                // filter out directions that would go outside the field
                let target = self.head.move_to(*direction);
                (0..board.width).contains(&target.x) && (0..board.height).contains(&target.y)
            })
            .filter(|direction| {
                let target = self.head.move_to(*direction);
                // don't collide with any snake
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // filter out snakes that are in our squad if body collisions are allowed
                        !(game.ruleset.settings.squad.allow_body_collisions
                            && self.squad == snake.squad
                            && self.id != snake.id)
                    })
                    .flat_map(|snake| {
                        // get all coordinates of the snake body. The tail of the body can be ignored
                        // if the snake hasn't just eaten, as it will move out of the way
                        let has_eaten = snake.health == MAX_HEALTH;
                        snake.body[..snake.body.len() - usize::from(!has_eaten)].iter()
                    })
                    .any(|&coord| coord == target)
            })
    }

    #[must_use]
    pub fn possible_actions(&self, game: &Game, board: &Board) -> Vec<Direction> {
        self.possible_actions_without_heads(game, board)
            .filter(|direction| {
                // don't go into spots where a bigger snake must go
                let target = self.head.move_to(*direction);
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // The other snake must be bigger than we are
                        snake.length > self.length
                    })
                    .filter_map(|snake| {
                        // get all snakes movement options
                        let actions = snake
                            .possible_actions_without_heads(game, board)
                            .collect::<Vec<_>>();
                        // only snakes that have a single option
                        match actions[..] {
                            [direction] => Some(snake.head.move_to(direction)),
                            _ => None,
                        }
                    })
                    .any(|coord| coord == target)
            })
            .collect()
    }
}

// info is called when you create your Battlesnake on play.battlesnake.com
// and controls your Battlesnake's appearance
// TIP: If you open your Battlesnake URL in a browser you should see this data
pub fn info() -> Value {
    info!("INFO");

    json!({
        "apiversion": "1",
        "author": "der-informatiker",
        "color": "#00FFEE",
        "head": "smart-caterpillar",
        "tail": "mouse",
    })
}

// start is called when your Battlesnake begins a game
pub fn start(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME START");
}

// end is called when your Battlesnake finishes a game
pub fn end(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME OVER");
}

// move is called on every turn and returns your next move
// Valid moves are "up", "down", "left", or "right"
// See https://docs.battlesnake.com/api/example-move for available data
pub fn get_move(game: &Game, turn: i32, board: &Board, you: &Battlesnake) -> Option<Action> {
    let token_map = SnakeToken::from_board(board);
    let board = simulation::Board::from_game_board(
        board,
        &token_map,
        turn,
        game.ruleset.settings.food_spawn_chance,
        game.ruleset.settings.minimum_food,
    );

    let my_token = token_map[&you.id];

    let mut tree = Node::default();

    for _ in 0..300 {
        let mut board = board.clone();
        tree.monte_carlo_step(&mut board);
    }

    let actions = tree.child_statistics.entry(my_token).or_default();

    info!("actions: {actions:?}");

    let chosen = actions
        .iter()
        .max_by_key(|(_, stat)| OrderedFloat(stat.won as f64 / stat.played as f64))
        .map(|(direction, _)| *direction)?;

    info!("DIRECTION {}: {:?}", turn, chosen);
    Some(Action {
        r#move: chosen,
        shout: None,
    })
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct Statistics {
    /// Number of times this node was simulated
    played: usize,
    /// Number of times this node was simulated and the agent has won.
    won: BTreeMap<SnakeToken, usize>,
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct ActionStatistic {
    played: usize,
    won: usize,
}

#[derive(Debug, PartialEq, Eq, Clone, Default)]
struct Node {
    statistic: Statistics,
    child_statistics: BTreeMap<SnakeToken, BTreeMap<Direction, ActionStatistic>>,
    childs: BTreeMap<BTreeMap<SnakeToken, Direction>, Node>,
}

impl Node {
    /// Performs one monte carlo simulation step
    ///
    /// Returns the snake that has won the simulation
    fn monte_carlo_step(&mut self, board: &mut simulation::Board) -> Option<SnakeToken> {
        let winner = if self.statistic.played == 0 {
            // We didn't simulate a game for this node yet. Do that
            board.simulate_until(|board| board.alive_snakes() <= 1);
            board.snakes().next()
        } else {
            // select a node to simulate
            let possible_actions = board.possible_actions();

            let actions = possible_actions
                .iter()
                .map(|(token, actions)| {
                    let statistics = self.child_statistics.entry(*token).or_default();
                    let selected = actions
                        .iter()
                        .copied()
                        .max_by_key(|direction| {
                            let statistics = statistics.entry(*direction).or_default();
                            if statistics.played == 0 {
                                return OrderedFloat(f64::INFINITY);
                            }
                            let exploitation = statistics.won as f64 / statistics.played as f64;
                            let exploration = f64::consts::SQRT_2
                                * f64::sqrt(
                                    f64::ln(self.statistic.played as f64)
                                        / statistics.played as f64,
                                );
                            OrderedFloat(exploitation + exploration)
                        })
                        .unwrap_or_default();
                    (*token, selected)
                })
                .collect();

            board.simulate_actions(&actions);
            let winner = self
                .childs
                .entry(actions.clone())
                .or_default()
                .monte_carlo_step(board);

            // update child statistics
            for (token, action) in &actions {
                let entry = self
                    .child_statistics
                    .entry(*token)
                    .or_default()
                    .entry(*action)
                    .or_default();
                entry.played += 1;
                if Some(*token) == winner {
                    entry.won += 1;
                }
            }

            winner
        };
        self.statistic.played += 1;
        if let Some(token) = winner {
            self.statistic
                .won
                .entry(token)
                .and_modify(|won| *won += 1)
                .or_insert(1);
        }
        winner
    }
}
